@@ -3,43 +3,44 @@
 %%
 %%     Various utility functions that not obviously fit somewhere else.
 %%
-%%  Copyright (c) 2001-2008 Bjorn Gustavsson
+%%  Copyright (c) 2001-2011 Bjorn Gustavsson
 %%
 %%  See the file "license.terms" for information on usage and redistribution
 %%  of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 %%
-%%     $Id: wings_util.erl 361 2008-07-22 15:46:15Z optigon $
+%%     $Id$
 %%
 %% Note: To keep the call graph clean, wings_util MUST NOT call
 %%       other wings_* modules (except wings_pref).
 
 -module(wings_util).
--export([wings/0,share/1,share/3,make_vector/1,
+-export([share/1,share/3,make_vector/1,
 	 rel2fam/1,
 	 format/2,
 	 key_format/2,
 	 cap/1,upper/1,stringify/1,quote/1,
-	 expand_utf8/1,
 	 add_vpos/2,update_vpos/2,
 	 gb_trees_smallest_key/1,gb_trees_largest_key/1,
-	 gb_trees_map/2,gb_trees_to_gb_set/1,
-	 nice_float/1,
+	 array_keys/1,array_smallest_key/1,array_greatest_key/1,
+	 array_is_empty/1,array_entries/1,
+	 mapsfind/3,
+	 wxequal/2, wxset_pid/2, min_wx/1,
+	 nice_float/1,nice_vector/1,nice_abs_vector/1,
+	 nonzero/1,
+	 string_to_float/1,
 	 unique_name/2,
+	 is_name_masked/2,
 	 lib_dir/1,
-	 tc/3,
-	 min/2,max/2,limit/2]).
+	 tc/3, profile_start/1, profile_stop/1,
+	 limit/2]).
 
 -define(NEED_OPENGL, 1).
 -define(NEED_ESDL, 1).
 -include("wings.hrl").
--import(lists, [foreach/2,map/2,foldl/3,reverse/1,member/2,last/1]).
+-include_lib("wings/e3d/e3d.hrl").
+
+-import(lists, [foldl/3,reverse/1,member/2,last/1]).
 
--ifdef(DEBUG).
-wings() -> "Wings 3D [debug]".
--else.
-wings() -> "Wings 3D".
--endif.
-    
 share(X, X, X) -> {X,X,X};
 share(X, X, Z) -> {X,X,Z};
 share(X, Y, Y) -> {X,Y,Y};
@@ -78,15 +79,18 @@ make_vector(z) -> {0.0,0.0,1.0};
 make_vector(free) -> free;
 make_vector(normal) -> normal;
 make_vector(intrude) -> normal;
+make_vector({region,normal}) -> normal;
 make_vector(Axis) when Axis == last_axis; Axis == default_axis ->
     {_,Vec} = wings_pref:get_value(Axis),
     Vec.
 
+
 key_format(Key, Msg) ->
     [Key,160,Msg].
 
 %% Like io_lib:format/2, but with very restricted format characters.
 %% BUT allows arguments to ~s to be lists containing Unicode characters.
+%% Now allows ~ts for translation of Asian glyphs.
 %%
 %% Format directives allowed: ~s ~p
 
@@ -129,104 +133,77 @@ upper([H|T]) ->
     [H|upper(T)];
 upper([]) -> [].
 
-%% expand_utf8([Byte]) -> {[UnicodeChar],NumberOfBadBytes}
-%%  Expand UTF8 byte sequences to ISO 10646/Unicode
-%%  charactes. Any illegal bytes are removed and the number of
-%%  bad bytes are returned.
-%%
-%%  Reference:
-%%     RFC 3629: "UTF-8, a transformation format of ISO 10646".
-
-expand_utf8(Str) ->
-    expand_utf8_1(Str, [], 0).
-
-expand_utf8_1([C|Cs], Acc, Bad) when C < 16#80 ->
-    %% Plain Ascii character.
-    expand_utf8_1(Cs, [C|Acc], Bad);
-expand_utf8_1([C1,C2|Cs], Acc, Bad) when C1 band 16#E0 =:= 16#C0,
-					 C2 band 16#C0 =:= 16#80 ->
-    case ((C1 band 16#1F) bsl 6) bor (C2 band 16#3F) of
-	C when 16#80 =< C ->
-	    expand_utf8_1(Cs, [C|Acc], Bad);
-	_ ->
-	    %% Bad range.
-	    expand_utf8_1(Cs, Acc, Bad+1)
-    end;
-expand_utf8_1([C1,C2,C3|Cs], Acc, Bad) when C1 band 16#F0 =:= 16#E0,
-					    C2 band 16#C0 =:= 16#80,
-					    C3 band 16#C0 =:= 16#80 ->
-    case ((((C1 band 16#0F) bsl 6) bor (C2 band 16#3F)) bsl 6) bor
-	(C3 band 16#3F) of
-	C when 16#800 =< C ->
-	    expand_utf8_1(Cs, [C|Acc], Bad);
-	_ ->
-	    %% Bad range.
-	    expand_utf8_1(Cs, Acc, Bad+1)
-    end;
-expand_utf8_1([C1,C2,C3,C4|Cs], Acc, Bad) when C1 band 16#F8 =:= 16#F0,
-					       C2 band 16#C0 =:= 16#80,
-					       C3 band 16#C0 =:= 16#80,
-					       C4 band 16#C0 =:= 16#80 ->
-    case ((((((C1 band 16#0F) bsl 6) bor (C2 band 16#3F)) bsl 6) bor
-	(C3 band 16#3F)) bsl 6) bor (C4 band 16#3F) of
-	C when 16#10000 =< C ->
-	    expand_utf8_1(Cs, [C|Acc], Bad);
-	_ ->
-	    %% Bad range.
-	    expand_utf8_1(Cs, Acc, Bad+1)
-    end;
-expand_utf8_1([_|Cs], Acc, Bad) ->
-    %% Ignore bad character.
-    expand_utf8_1(Cs, Acc, Bad+1);
-expand_utf8_1([], Acc, Bad) -> {reverse(Acc),Bad}.
-    
-
 add_vpos(Vs, #we{vp=Vtab}) -> add_vpos(Vs, Vtab);
 add_vpos(Vs, Vtab) ->
     foldl(fun(V, A) ->
-		  [{V,gb_trees:get(V, Vtab)}|A]
+		  [{V,array:get(V, Vtab)}|A]
 	  end, [], Vs).
 
 update_vpos(Vs, #we{vp=Vtab}) -> update_vpos(Vs, Vtab);
 update_vpos(Vs, Vtab) ->
     foldl(fun({V,_}, A) ->
-		  [{V,gb_trees:get(V, Vtab)}|A];
+		  [{V,array:get(V, Vtab)}|A];
 	     ({V,_,Dist,Inf}, A) ->
-		  [{V,gb_trees:get(V, Vtab),Dist,Inf}|A]
+		  [{V,array:get(V, Vtab),Dist,Inf}|A]
 	  end, [], reverse(Vs)).
 
-gb_trees_smallest_key({_, Tree}) ->
-    smallest_key1(Tree).
+gb_trees_smallest_key(Tree) ->
+    {Key,_Val} = gb_trees:smallest(Tree),
+    Key.
 
-smallest_key1({Key, _Value, nil, _Larger}) ->
-    Key;
-smallest_key1({_Key, _Value, Smaller, _Larger}) ->
-    smallest_key1(Smaller).
+gb_trees_largest_key(Tree) ->
+    {Key,_Val} = gb_trees:largest(Tree),
+    Key.
 
-gb_trees_largest_key({_, Tree}) ->
-    largest_key1(Tree).
+array_keys(Array) ->
+    array:sparse_foldr(fun(I, _, A) -> [I|A] end, [], Array).
 
-largest_key1({Key, _Value, _Smaller, nil}) ->
-    Key;
-largest_key1({_Key, _Value, _Smaller, Larger}) ->
-    largest_key1(Larger).
+array_smallest_key(Array) ->
+    try
+	array:sparse_foldl(fun(I, _, _) -> throw(I) end, [], Array),
+	error(empty_array)
+    catch
+	throw:I when is_integer(I) ->
+	    I
+    end.
 
-gb_trees_map(F, {Size,Tree}) ->
-    {Size,gb_trees_map_1(F, Tree)}.
+array_greatest_key(Array) ->
+    try
+	array:sparse_foldr(fun(I, _, _) -> throw(I) end, [], Array),
+	error(empty_array)
+    catch
+	throw:I when is_integer(I) ->
+	    I
+    end.
 
-gb_trees_map_1(_, nil) -> nil;
-gb_trees_map_1(F, {K,V,Smaller,Larger}) ->
-    {K,F(K, V),
-     gb_trees_map_1(F, Smaller),
-     gb_trees_map_1(F, Larger)}.
+array_is_empty(Array) ->
+    try
+	array:sparse_foldr(fun(_, _, _) -> throw(false) end, [], Array),
+	throw(true)
+    catch
+	throw:Empty ->
+	    Empty
+    end.
 
-gb_trees_to_gb_set({Size,Tree}) ->
-    {Size,gb_trees_to_gb_set_1(Tree)}.
+%% array_entries(Array) -> NumberOfEntries
+%%  Return the number of non-default entries in the array.
+%%
+array_entries(Array) ->
+    array:sparse_foldl(fun(_, _, N) -> N + 1 end, 0, Array).
 
-gb_trees_to_gb_set_1(nil) ->
-    nil;
-gb_trees_to_gb_set_1({K,_,Smaller,Larger}) ->
-    {K,gb_trees_to_gb_set_1(Smaller),gb_trees_to_gb_set_1(Larger)}.
+-spec nice_abs_vector(e3d_vec:vector()) -> iolist().
+
+nice_abs_vector({X,Y,Z}) ->
+    nice_vector({abs(X),abs(Y),abs(Z)}).
+
+-spec nice_vector(e3d_vec:vector()) -> iolist().
+
+nice_vector({X,Y,Z}) ->
+    ["<",
+     wings_util:nice_float(X),"  ",
+     wings_util:nice_float(Y),"  ",
+     wings_util:nice_float(Z),
+     ">"].
 
 nice_float(F) when is_float(F) ->
     simplify_float(lists:flatten(io_lib:format("~f", [F]))).
@@ -238,6 +215,68 @@ simplify_float_1("0."++_=F) -> F;
 simplify_float_1("0"++F) -> simplify_float_1(F);
 simplify_float_1(F) -> F.
 
+nonzero(Value) ->
+    if abs(Value) < ?EPSILON -> ?EPSILON;
+    true -> Value
+    end.
+
+string_to_float("NaN") -> 0.0;
+string_to_float(Str) ->
+    try list_to_float(Str)
+    catch _:_ -> make_float2(Str)
+    end.
+
+make_float2(Str) ->
+    try float(list_to_integer(Str))
+    catch _:_ -> make_float3(Str)
+    end.
+
+make_float3(Str0) ->
+    Str = lists:flatten(string:tokens(Str0, "\t\s\n")),
+    try list_to_float(Str)
+    catch _:_ -> make_float4(Str) end.
+
+make_float4(Str) ->
+    WithDot = case string:tokens(Str, "eE") of
+		  [Pre,Post] -> Pre ++ ".0e" ++ Post;
+		  [Other] -> Other ++ ".0";
+		  Other -> Other
+	      end,
+    list_to_float(WithDot).
+
+%%
+%% Finds the first map containing Key:=Value
+%%
+-spec mapsfind(term(), term(), list()) -> map()|false.
+mapsfind(Value, Key, [H|T]) ->
+    case H of
+	#{Key:=Value} -> H;
+	_ -> mapsfind(Value, Key, T)
+    end;
+mapsfind(_, _, []) -> false.
+
+
+%% Missing wx funcs
+-record(wx_ref, {ref, type, state}).
+wxequal(#wx_ref{ref=Ref1}, #wx_ref{ref=Ref2}) -> Ref1 =:= Ref2.
+wxset_pid(#wx_ref{}=R, Pid) when is_pid(Pid) ->
+    R#wx_ref{state=Pid}.
+
+min_wx({_,_}=Ver) ->
+    case get(wx_version) of
+        undefined ->
+            case filename:basename(code:lib_dir(wx)) of
+                [$w,$x,$-,Major,_,Minor|_] ->
+                    Version = {Major-$0, Minor-$0},
+                    put(wx_version, Version),
+                    Version >= Ver;
+                _Vsn -> %% erlang src build? assume 19.2 or later
+                    true
+            end;
+        Version ->
+            Version >= Ver
+    end.
+
 %%
 %% Create a unique name by appending digits.
 %%
@@ -267,18 +306,38 @@ unique_name_2(Base, I, Names) ->
     end.
 
 tc(Fun,Mod,Line) ->
-    case timer:tc(erlang, apply, [Fun,[]]) of
-	{_,{'EXIT',Reason}} -> exit(Reason);
-	{T,R} ->
-	    io:format("~p:~p: Time: ~p\n", [Mod, Line, T]),
-	    R
-    end.
+    Before = os:timestamp(),
+    R = Fun(),
+    After = os:timestamp(),
+    io:format("~p:~p: Time: ~p\n", [Mod, Line, timer:now_diff(After,Before)]),
+    R.
 
-max(A, B) when A > B -> A;
-max(_A, B) -> B.
+profile_start(fprof) ->
+    fprof:trace(start),
+    ok;
+profile_start(eprof) ->
+    eprof:start(),
+    profiling = eprof:start_profiling([whereis(wings), self(), whereis(wings_image)]),
+    ok.
 
-min(A, B) when A < B -> A;
-min(_A, B) -> B.
+profile_stop(fprof) ->
+    fprof:trace(stop),
+    spawn_link(fun() ->
+                       File = "fprof.analysis",
+                       fprof:profile(),
+                       fprof:analyse([{dest, File}, {cols, 120}]),
+                       io:format("Analysis in: ~p~n", [filename:absname(File)]),
+                       eprof:stop()
+               end),
+    ok;
+profile_stop(eprof) ->
+    eprof:stop_profiling(),
+    spawn_link(fun() ->
+                       File = "eprof.analysis",
+                       eprof:log(File),
+                       eprof:analyze(),
+                       io:format("Analysis in: ~p~n", [filename:absname(File)])
+               end).
 
 limit(Val, {'-infinity',infinity}) -> Val;
 limit(Val, {Min,infinity}) when Val < Min -> Min;
@@ -287,6 +346,59 @@ limit(Val, {Min,Max}) when Min < Max, Val < Min -> Min;
 limit(Val, {Min,Max}) when Min < Max, Val > Max -> Max;
 limit(Val, {Min,Max}) when Min < Max -> Val.
 
+%%
+%% Check if name match with the mask.
+%%
+
+is_name_masked(Name,Mask) ->
+    Mask0=string:to_upper(Mask),
+    Name0=string:to_upper(Name),
+    is_name_masked_0(Name0,Mask0,get_mask_pos(Mask0)).
+
+is_name_masked_0(_Name,"*",_MaskPos) -> true;
+is_name_masked_0([],_Mask,_MaskPos) -> false;
+is_name_masked_0(_Name,[],_MaskPos) -> false;
+is_name_masked_0(Name,Name,[]) -> true;
+is_name_masked_0(_Name,_Mask,[]) -> false;
+is_name_masked_0(Name,Mask,[H|[]]=_MaskPos) ->  % *text with only one wildcard
+    Mlen=string:len(Mask),
+    case H of
+    1 ->     % *text value
+        Mask0=string:sub_string(Mask,H+1),
+        MPos=string:rstr(Name,Mask0),
+        (string:len(Name)-MPos+1)=:=(Mlen-1);
+    Mlen ->  % text value*
+        Mask0=string:sub_string(Mask,1,Mlen-1),
+        string:str(Name,Mask0)=:=1;
+    _ ->     % text *value
+        Mask0=string:sub_string(Mask,1,H-1),
+        Mask1=string:sub_string(Mask,H+1),
+        MPos=string:rstr(Name,Mask1),
+        Mlen0=string:len(Mask1),
+        (string:str(Name,Mask0)=:=1) and ((string:len(Name)-MPos)=:=(Mlen0-1))
+    end;
+is_name_masked_0(Name,Mask,[H|[H0|_T0]=_T]=_MaskPos) ->  % *text with more than one wildcard
+    case H of
+    1 ->     % *text value
+        Mask0=string:sub_string(Mask,H+1,H0-1),
+        Mask1=string:sub_string(Mask,H0),
+        case string:str(Name,Mask0) of
+        0 -> false;
+        NPos ->
+            Name0=string:sub_string(Name,NPos+H0-H-1),
+            is_name_masked_0(Name0,Mask1,get_mask_pos(Mask1))
+        end;
+    _ ->     % te*xt value*
+        Mask0=string:sub_string(Mask,1,H-1),
+        Mask1=string:sub_string(Mask,H),
+        case string:str(Name,Mask0) of
+        1 ->
+            Name0=string:sub_string(Name,H),
+            is_name_masked_0(Name0,Mask1,get_mask_pos(Mask1));
+        _ -> false
+        end
+    end.
+
 %%%
 %%% Local functions.
 %%%
@@ -298,7 +410,7 @@ format_1("~s"++F, [S0|Args], Acc) ->
 	    true -> 
 		io:format("Bad string formatter for ~p in ~p~n", 
 			  [S0, lists:flatten(reverse(Acc) ++ F)]),
-		erlang:error(badarg)
+		error(badarg)
 	end,
     format_1(F, Args, [S|Acc]);
 format_1("~p"++F, [S0|Args], Acc) ->
@@ -306,6 +418,8 @@ format_1("~p"++F, [S0|Args], Acc) ->
     format_1(F, Args, [S|Acc]);
 format_1("~~"++F, Args, Acc) ->
     format_1(F, Args, [$~|Acc]);
+format_1("~ts"++F, Args, Acc) ->
+    format_1("~s"++F, Args, Acc);
 format_1([C|F], Args, Acc) when C =/= $~ ->
     format_1(F, Args, [C|Acc]);
 format_1([], [], Acc) -> reverse(Acc).
@@ -326,6 +440,20 @@ lib_dir(wings) ->
 lib_dir(Lib) ->
     code:lib_dir(Lib).
 
+get_mask_pos([]) -> [];
+get_mask_pos(Mask) ->
+    get_mask_pos_1(Mask,0,[]).
+
+get_mask_pos_1([],_,Acc) -> Acc;
+get_mask_pos_1(Mask,Offset,Acc) ->
+    case string:str(Mask,"*") of
+    0 -> Acc;
+    Pos ->
+        Pos0=Pos+Offset,
+        get_mask_pos_1(string:sub_string(Mask,Pos+1),Pos0,Acc++[Pos0])
+    end.
+
+
 % % % % % % % % % % % % % % % % % % %
 %% Strings for Translating Hotkeys %%
 % % % % % % (not elegant) % % % % % %
@@ -492,7 +620,7 @@ translation_string(outliner)         -> ?__(137,"Outliner");
 translation_string(object)           -> ?__(138,"Geometry Graph");
 translation_string(palette)          -> ?__(139,"Palette");
 translation_string(console)          -> ?__(140,"Console");
-translation_string(geom_viewer)      -> ?__(141,"New Geomerty Window");
+translation_string(geom_viewer)      -> ?__(141,"New Geometry Window");
 translation_string(uv_editor_window) -> ?__(142,"UV Editor Window");
 
 %%%% Help Menu
@@ -501,7 +629,7 @@ translation_string(one_or_two)           -> ?__(144,"Mouse Buttons");
 translation_string(international)        -> ?__(145,"International Keyboards");
 translation_string(hotkeys)              -> ?__(146,"Defined Hotkeys");
 translation_string(defining_hotkeys)     -> ?__(147,"Defining Hotkeys");
-translation_string(advanced_menus)       -> ?__(148,"Advanced Menus");
+%%translation_string(advanced_menus)       -> ?__(148,"Advanced Menus");
 translation_string(default_commands)     -> ?__(149,"Default Commands");
 translation_string(performance_tips)     -> ?__(150,"Performance Tips");
 translation_string(opengl_info)          -> ?__(151,"OpenGL Info");
@@ -670,11 +798,6 @@ translation_string(materials_to_colors)  -> ?__(288,"Materials to Colors");
 %%%%   Translation strings used so far 1 - 290
 %%%%
 
-%%%% Others as yet to be added are proccessed here
+%%%% Others as yet to be added are processed here
 translation_string(Atom) when is_atom(Atom) ->
-    wings_util:cap(atom_to_list(Atom));
-
-translation_string(Int) when is_integer(Int) ->
-    integer_to_list(Int);
-
-translation_string(_Other) -> [].
+    wings_util:cap(atom_to_list(Atom)).
